@@ -15,14 +15,14 @@ if gevent_socketio_found:
 
 import socketio
 import flask
-from flask import json as flask_json
+from flask import _request_ctx_stack, json as flask_json
 from werkzeug.debug import DebuggedApplication
 from werkzeug.serving import run_with_reloader
 
 from .namespace import Namespace
 from .test_client import SocketIOTestClient
 
-__version__ = '2.8.2'
+__version__ = '2.9.2'
 
 
 class _SocketIOMiddleware(socketio.Middleware):
@@ -49,6 +49,14 @@ class SocketIO(object):
                 isn't known at the time this class is instantiated, then call
                 ``socketio.init_app(app)`` once the application instance is
                 available.
+    :param manage_session: If set to ``True``, this extension manages the user
+                           session for Socket.IO events. If set to ``False``,
+                           Flask's own session management is used. When using
+                           Flask's cookie based sessions it is recommended that
+                           you leave this set to the default of ``True``. When
+                           using server-side sessions, a ``False`` setting
+                           enables sharing the user session between HTTP routes
+                           and Socket.IO events.
     :param message_queue: A connection URL for a message queue service the
                           server can use for multi-process communication. A
                           message queue is not required when using a single
@@ -128,6 +136,7 @@ class SocketIO(object):
         self.namespace_handlers = []
         self.exception_handlers = {}
         self.default_exception_handler = None
+        self.manage_session = True
         # We can call init_app when:
         # - we were given the Flask app instance (standard initialization)
         # - we were not given the app, but we were given a message_queue
@@ -145,14 +154,18 @@ class SocketIO(object):
                 app.extensions = {}  # pragma: no cover
             app.extensions['socketio'] = self
         self.server_options.update(kwargs)
+        self.manage_session = self.server_options.pop('manage_session',
+                                                      self.manage_session)
 
         if 'client_manager' not in self.server_options:
-            url = kwargs.pop('message_queue', None)
-            channel = kwargs.pop('channel', 'flask-socketio')
+            url = self.server_options.pop('message_queue', None)
+            channel = self.server_options.pop('channel', 'flask-socketio')
             write_only = app is None
             if url:
                 if url.startswith('redis://'):
                     queue_class = socketio.RedisManager
+                elif url.startswith('zmq'):
+                    queue_class = socketio.ZmqManager
                 else:
                     queue_class = socketio.KombuManager
                 queue = queue_class(url, channel=channel,
@@ -178,7 +191,8 @@ class SocketIO(object):
 
             self.server_options['json'] = FlaskSafeJSON
 
-        resource = kwargs.pop('path', kwargs.pop('resource', 'socket.io'))
+        resource = self.server_options.pop('path', None) or \
+            self.server_options.pop('resource', None) or 'socket.io'
         if resource.startswith('/'):
             resource = resource[1:]
         self.server = socketio.Server(**self.server_options)
@@ -225,7 +239,7 @@ class SocketIO(object):
                 self.server.on(message, _handler, namespace=namespace)
             else:
                 self.handlers.append((message, _handler, namespace))
-            return _handler
+            return handler
         return decorator
 
     def on_error(self, namespace=None):
@@ -469,7 +483,11 @@ class SocketIO(object):
             def run_server():
                 import eventlet
                 import eventlet.wsgi
-                eventlet_socket = eventlet.listen((host, port))
+                import eventlet.green
+                addresses = eventlet.green.socket.getaddrinfo(host, port)
+                if not addresses:
+                    raise RuntimeError('Could not resolve host to a valid address')
+                eventlet_socket = eventlet.listen(addresses[0][4], addresses[0][0])
 
                 # If provided an SSL argument, use an SSL socket
                 ssl_args = ['keyfile', 'certfile', 'server_side', 'cert_reqs',
@@ -508,7 +526,7 @@ class SocketIO(object):
                     log=log, **kwargs)
             else:
                 self.wsgi_server = pywsgi.WSGIServer((host, port), app,
-                                                     log=log)
+                                                     log=log, **kwargs)
 
             if use_reloader:
                 # monkey patching is required by the reloader
@@ -565,9 +583,10 @@ class SocketIO(object):
         """
         return self.server.sleep(seconds)
 
-    def test_client(self, app, namespace=None):
+    def test_client(self, app, namespace=None, query_string=None, headers=None):
         """Return a simple SocketIO client that can be used for unit tests."""
-        return SocketIOTestClient(app, self, namespace)
+        return SocketIOTestClient(app, self, namespace=namespace,
+                                  query_string=query_string, headers=headers)
 
     def _handle_event(self, handler, message, namespace, sid, *args):
         if sid not in self.server.environ:
@@ -575,10 +594,21 @@ class SocketIO(object):
             return '', 400
         app = self.server.environ[sid]['flask.app']
         with app.request_context(self.server.environ[sid]):
-            if 'saved_session' in self.server.environ[sid]:
-                self._copy_session(
-                    self.server.environ[sid]['saved_session'],
-                    flask.session)
+            if self.manage_session:
+                # manage a separate session for this client's Socket.IO events
+                # created as a copy of the regular user session
+                if 'saved_session' not in self.server.environ[sid]:
+                    self.server.environ[sid]['saved_session'] = \
+                        dict(flask.session)
+                session_obj = self.server.environ[sid]['saved_session']
+            else:
+                # let Flask handle the user session
+                # for cookie based sessions, this effectively freezes the
+                # session to its state at connection time
+                # for server-side sessions, this allows HTTP and Socket.IO to
+                # share the session, with both having read/write access to it
+                session_obj = flask.session._get_current_object()
+            _request_ctx_stack.top.session = session_obj
             flask.request.sid = sid
             flask.request.namespace = namespace
             flask.request.event = {'message': message, 'args': args}
@@ -594,17 +624,13 @@ class SocketIO(object):
                     raise
                 type, value, traceback = sys.exc_info()
                 return err_handler(value)
-            if flask.session.modified and sid in self.server.environ:
-                self.server.environ[sid]['saved_session'] = {}
-                self._copy_session(
-                    flask.session,
-                    self.server.environ[sid]['saved_session'])
+            if not self.manage_session:
+                # when Flask is managing the user session, it needs to save it
+                if not hasattr(session_obj, 'modified') or session_obj.modified:
+                    resp = app.response_class()
+                    app.session_interface.save_session(app, session_obj, resp)
             return ret
 
-    def _copy_session(self, src, dest):
-        for k in src:
-            dest[k] = src[k]
-
 
 def emit(event, *args, **kwargs):
     """Emit a SocketIO event.
@@ -633,6 +659,13 @@ def emit(event, *args, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addresee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
     if 'namespace' in kwargs:
         namespace = kwargs['namespace']
@@ -644,10 +677,12 @@ def emit(event, *args, **kwargs):
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
     return socketio.emit(event, *args, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+                         include_self=include_self, callback=callback,
+                         ignore_queue=ignore_queue)
 
 
 def send(message, **kwargs):
@@ -671,6 +706,13 @@ def send(message, **kwargs):
     :param include_self: ``True`` to include the sender when broadcasting or
                          addressing a room, or ``False`` to send to everyone
                          but the sender.
+    :param ignore_queue: Only used when a message queue is configured. If
+                         set to ``True``, the event is emitted to the
+                         clients directly, without going through the queue.
+                         This is more efficient, but only works when a
+                         single server process is used, or when there is a
+                         single addresee. It is recommended to always leave
+                         this parameter with its default value of ``False``.
     """
     if 'namespace' in kwargs:
         namespace = kwargs['namespace']
@@ -682,13 +724,15 @@ def send(message, **kwargs):
     if room is None and not broadcast:
         room = flask.request.sid
     include_self = kwargs.get('include_self', True)
+    ignore_queue = kwargs.get('ignore_queue', False)
 
     socketio = flask.current_app.extensions['socketio']
     return socketio.send(message, namespace=namespace, room=room,
-                         include_self=include_self, callback=callback)
+                         include_self=include_self, callback=callback,
+                         ignore_queue=ignore_queue)
 
 
-def join_room(room):
+def join_room(room, sid=None, namespace=None):
     """Join a room.
 
     This function puts the user in a room, under the current namespace. The
@@ -703,18 +747,22 @@ def join_room(room):
             send(username + ' has entered the room.', room=room)
 
     :param room: The name of the room to join.
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.enter_room(flask.request.sid, room,
-                               namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    socketio.server.enter_room(sid, room, namespace=namespace)
 
 
-def leave_room(room):
+def leave_room(room, sid=None, namespace=None):
     """Leave a room.
 
     This function removes the user from a room, under the current namespace.
-    The user and the namespace are obtained from the event context. This is
-    a function that can only be called from a SocketIO event handler. Example::
+    The user and the namespace are obtained from the event context. Example::
 
         @socketio.on('leave')
         def on_leave(data):
@@ -724,35 +772,47 @@ def leave_room(room):
             send(username + ' has left the room.', room=room)
 
     :param room: The name of the room to leave.
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.leave_room(flask.request.sid, room,
-                               namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    socketio.server.leave_room(sid, room, namespace=namespace)
 
 
-def close_room(room):
+def close_room(room, namespace=None):
     """Close a room.
 
     This function removes any users that are in the given room and then deletes
-    the room from the server. This is a function that can only be called from
-    a SocketIO event handler.
+    the room from the server.
 
     :param room: The name of the room to close.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    socketio.server.close_room(room, namespace=flask.request.namespace)
+    namespace = namespace or flask.request.namespace
+    socketio.server.close_room(room, namespace=namespace)
 
 
-def rooms():
+def rooms(sid=None, namespace=None):
     """Return a list of the rooms the client is in.
 
     This function returns all the rooms the client has entered, including its
-    own room, assigned by the Socket.IO server. This is a function that can
-    only be called from a SocketIO event handler.
+    own room, assigned by the Socket.IO server.
+
+    :param sid: The session id of the client. If not provided, the client is
+                obtained from the request context.
+    :param namespace: The namespace for the room. If not provided, the
+                      namespace is obtained from the request context.
     """
     socketio = flask.current_app.extensions['socketio']
-    return socketio.server.rooms(flask.request.sid,
-                                 namespace=flask.request.namespace)
+    sid = sid or flask.request.sid
+    namespace = namespace or flask.request.namespace
+    return socketio.server.rooms(sid, namespace=namespace)
 
 
 def disconnect(silent=False):
